"""
Train_logger
============
"""
from torch.utils.tensorboard import SummaryWriter
from conformer_rl.utils import mkdir

class TrainLogger:
    """Used by agent for logging agent metrics during training.

    Parameters
    ----------
    tag : str
        Unique tag for identifying the logging session.
    dir : str
        Path to root directory for where logging results should be saved.
    use_tensorboard : bool
        Whether or not to save metrics to Tensorboard.
    use_cache : bool
        Whether or not to keep a running cache of the metrics.
    use_print : bool
        Whether or not to print metrics when logged.
    """
    def __init__(self, tag: str, dir: str = "data", use_tensorboard: bool = True, use_cache: bool = True, use_print: bool = True):
        self.dir = dir
        self.tag = tag
        mkdir(dir)
        self.use_tensorboard = use_tensorboard
        self.use_cache = use_cache
        self.use_print = use_print

        if self.use_tensorboard:
            path = dir + '/' + 'tensorboard_log' + '/' + self.tag + '/'
            mkdir(path)
            self.writer = SummaryWriter(log_dir = path)

        if self.use_cache:
            self.cache = {}


    def add_scalar(self, key: str, scalar_value: float, global_step: int=None, walltime: float=None) -> None:
        """Logs a single scalar value.

        Parameters
        ----------
        key : str
            The key associated with the logged value.
        scalar_value : float
            The value to be logged.
        global_step : int
            The current agent step when logging the metric.
        walltime : float
            The current time elapsed since the start of agent training.
        """
        if self.use_tensorboard:
            self.writer.add_scalar(key, scalar_value, global_step, walltime)

        if self.use_cache:
            if key in self.cache:
                self.cache[key][0].append(scalar_value)
                self.cache[key][1].append(global_step)
            else:
                self.cache[key] = [[scalar_value], [global_step]]

        if self.use_print:
            print("step:", global_step, key + ":", scalar_value)






"""
Molecules
=========
:class:`~conformer_rl.config.mol_config.MolConfig` generators.
"""
from conformer_rl.config import MolConfig
from conformer_rl.molecule_generation import generation
from conformer_rl.utils import calculate_normalizers
from rdkit import Chem
from rdkit.Chem import AllChem

def branched_alkane(num_atoms: int) -> MolConfig:
    """Generates a randomized branched alkane :class:`~conformer_rl.config.mol_config.MolConfig`,
    including constants for calculating Gibbs Score.

    Parameters
    ----------
    num_atoms : int
        The number of atoms in the branched alkane.
    """
    config = MolConfig()

    mol = generation.generate_branched_alkane(num_atoms)
    mol = Chem.AddHs(mol)
    AllChem.MMFFSanitizeMolecule(mol)
    E0, Z0 = calculate_normalizers(mol)
    
    config.mol = mol
    config.E0 = E0
    config.Z0 = Z0
    return config

def straight_alkane(num_atoms: int) -> MolConfig:
    """Generates a straight alkane chain :class:`~conformer_rl.config.mol_config.MolConfig`,
    including constants for calculating Gibbs Score.

    Parameters
    ----------
    num_atoms : int
        The number of atoms in the alkane.
    """
    config = MolConfig()

    mol = Chem.MolFromSmiles('C' * num_atoms)
    mol = Chem.AddHs(mol)
    AllChem.MMFFSanitizeMolecule(mol)
    E0, Z0 = calculate_normalizers(mol)

    config.mol = mol
    config.E0 = E0
    config.Z0 = Z0
    return config

def lignin(num_monomers: int) -> MolConfig:
    """Generates a lignin :class:`~conformer_rl.config.mol_config.MolConfig`,
    including constants for calculating Gibbs Score.

    Parameters
    ----------
    num_monomers : int
        Number of monomers in the lignin.
    """
    config = MolConfig()

    mol = generation.generate_lignin(num_monomers)
    mol = Chem.AddHs(mol)
    AllChem.MMFFSanitizeMolecule(mol)
    E0, Z0 = calculate_normalizers(mol)

    config.mol = mol
    config.E0 = E0
    config.Z0 = Z0
    return config

def xorgate(gate_complexity: int, num_gates: int) -> MolConfig:
    """Generates a xorgate :class:`~conformer_rl.config.mol_config.MolConfig`.

    An xorgate is a chain of alternating gates, where each gate is a chain of benzenes
    with a single carbon chain tail. 

    Parameters
    ----------
    gate_complexity : int
        Number of benzenes in each gate.
    num_gates : int
        Number of gates in molecule.
    """

    config = MolConfig()
    mol = generation.generate_xor_gate(gate_complexity, num_gates)
    mol = Chem.AddHs(mol)
    AllChem.MMFFSanitizeMolecule(mol)
    E0, Z0 = calculate_normalizers(mol)

    config.mol = mol
    config.E0 = E0
    config.Z0 = Z0
    return config



def test_alkane() -> MolConfig:
    """Generates a branched alkane :class:`~conformer_rl.config.mol_config.MolConfig` for testing.
    """
    config = MolConfig()
    
    mol = Chem.MolFromSmiles("CC(CCC)CCCC(CCCC)CC")
    mol = Chem.AddHs(mol)
    AllChem.MMFFSanitizeMolecule(mol)

    config.mol = mol
    config.E0 = 7.668625034772399
    config.Z0 = 13.263723987526067
    config.tau = 503
    return config

def mol_from_molFile(file: str, ep_steps: int = 200, pruning_thresh: float = 0.05) -> MolConfig:
    """
    """
    config = MolConfig()

    mol = Chem.MolFromMolFile(file)
    mol = _preprocess_mol(mol)
    config.mol = mol
    config.E0, config.Z0 = calculate_normalizers(mol, ep_steps, pruning_thresh)
    return config

def mol_from_smiles(smiles: str, ep_steps: int = 200, pruning_thresh: float  = 0.05) -> MolConfig:
    """
    """
    config = MolConfig()

    mol = Chem.MolFromSmiles(smiles)
    mol = _preprocess_mol(mol)
    config.mol = mol
    config.E0, config.Z0 = calculate_normalizers(mol, ep_steps, pruning_thresh)
    return config


def _preprocess_mol(mol: Chem.rdchem.Mol) -> Chem.rdchem.Mol:
    mol = Chem.AddHs(mol)
    AllChem.MMFFSanitizeMolecule(mol)

    return mol



"""
A2C_agent
=========
"""
import torch
import torch.nn as nn

from conformer_rl.agents.base_ac_agent import BaseACAgent
from conformer_rl.utils import to_np
from conformer_rl.agents.storage import Storage
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class A2CAgent(BaseACAgent):
    """ Implements agent that uses the A2C (advantage actor critic) [1]_ algorithm.

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by this agent.

    Notes
    -----
    *Config parameters:* The following parameters are required in the `config` object. See :class:`~conformer_rl.config.agent_config.Config`
    for more details on the parameters.

    * tag
    * train_env
    * eval_env
    * optimizer_fn
    * network
    * num_workers
    * rollout_length
    * max_steps
    * save_interval
    * eval_interval
    * eval_episodes
    * discount
    * use_gae
    * gae_lambda
    * entropy_weight
    * value_loss_coefficient
    * gradient_clip
    * ppo_ratio_clip
    * data_dir
    * use_tensorboard

    *Logged values*: The following values are logged during training:
    
    * advantages
    * loss
    * policy_loss
    * entropy_loss
    * value_loss
    * episodic_return_eval (total rewards per episode for eval episodes)
    * episodic_return_train (total rewards per episode for training episodes)


    References
    ----------
    .. [1] `A2C Paper <https://arxiv.org/abs/1602.01783>`_

    """
    def _train(self) -> None:
        storage = self.storage
        config = self.config

        actions = storage.order('a')
        log_prob = storage.order('log_pi_a')
        value = storage.order('v')
        entropy = storage.order('ent')
        returns = torch.stack(self.returns, 1).view(config.num_workers * config.rollout_length, -1)
        advantages = torch.stack(self.advantages, 1).view(config.num_workers * config.rollout_length, -1)

        entropy_loss = entropy.mean()
        policy_loss = -(log_prob * advantages).mean()
        value_loss = 0.5*(returns - value).pow(2).mean()
        loss = policy_loss - config.entropy_weight * entropy_loss + config.value_loss_weight * value_loss

        self.train_logger.add_scalar('policy_loss', policy_loss, self.total_steps)
        self.train_logger.add_scalar('value_loss', value_loss, self.total_steps)
        self.train_logger.add_scalar('entropy_loss', entropy_loss, self.total_steps)
        self.train_logger.add_scalar('loss', loss, self.total_steps)

        self.optimizer.zero_grad()
        loss.backward()
        grad_norm = nn.utils.clip_grad_norm_(self.network.parameters(), config.gradient_clip)
        self.train_logger.add_scalar('grad_norm', grad_norm, self.total_steps)
        self.optimizer.step()



"""
A2C_recurrent_agent
===================
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from conformer_rl.agents.base_ac_agent_recurrent import BaseACAgentRecurrent
from conformer_rl.utils import to_np
from conformer_rl.agents.storage import Storage
from conformer_rl.config import Config
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class A2CRecurrentAgent(BaseACAgentRecurrent):
    """ Implements agent that uses the A2C (advantage actor critic) [1]_ algorithm
    with support for recurrent neural networks.

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by this agent.

    Notes
    -----
    *Config parameters:* The following parameters are required in the `config` object. See :class:`~conformer_rl.config.agent_config.Config`
    for more details on the parameters.

    * tag
    * train_env
    * eval_env
    * optimizer_fn
    * network
    * num_workers
    * rollout_length
    * max_steps
    * save_interval
    * eval_interval
    * eval_episodes
    * recurrence
    * discount
    * use_gae
    * gae_lambda
    * entropy_weight
    * value_loss_coefficient
    * gradient_clip
    * ppo_ratio_clip
    * data_dir
    * use_tensorboard

    *Logged values*: The following values are logged during training:
    
    * advantages
    * loss
    * policy_loss
    * entropy_loss
    * value_loss
    * episodic_return_eval (total rewards per episode for eval episodes)
    * episodic_return_train (total rewards per episode for training episodes)

    References
    ----------
    .. [1] `A2C Paper <https://arxiv.org/abs/1602.01783>`_

    """
    def __init__(self, config: Config):
        super().__init__(config)
        assert config.rollout_length * config.num_workers % self.recurrence == 0

    def _train(self) -> None:
        config = self.config
        storage = self.storage

        actions = storage.order('a')
        returns = torch.stack(self.returns, 1).view(config.num_workers * config.rollout_length, -1)
        advantages = torch.stack(self.advantages, 1).view(config.num_workers * config.rollout_length, -1)

        recurrent_states = [storage.order(f'recurrent_states_{i}') for i in range(self.num_recurrent_units)]
        states = storage.order('states')

        total_entropy, total_value_loss, total_policy_loss, total_loss = 0, 0, 0, 0

        starting_indices = np.arange(0, self.config.rollout_length * self.config.num_workers, self.recurrence)
        sampled_recurrent_states = (recurrent_states[i][:, starting_indices]for i in range(self.num_recurrent_units))
        for i in range(self.recurrence):
            sampled_actions = actions[starting_indices + i]
            sampled_states = [states[j] for j in (starting_indices + i)]
            sampled_returns = returns[starting_indices + i]
            sampled_advantages = advantages[starting_indices + i]

            prediction, sampled_recurrent_states = self.network(sampled_states, sampled_recurrent_states, sampled_actions)
            entropy_loss = prediction['ent'].mean()
            policy_loss = -(prediction['log_pi_a'] * sampled_advantages).mean()
            value_loss = 0.5 * (sampled_returns - prediction['v']).pow(2).mean()

            loss = policy_loss - config.entropy_weight * entropy_loss + config.value_loss_weight * value_loss

            total_entropy += entropy_loss
            total_policy_loss += policy_loss
            total_value_loss += value_loss
            total_loss += loss


        total_entropy /= self.recurrence
        total_policy_loss /= self.recurrence
        total_value_loss /= self.recurrence
        total_loss /= self.recurrence

        self.train_logger.add_scalar('policy_loss', total_policy_loss, self.total_steps)
        self.train_logger.add_scalar('value_loss', total_value_loss, self.total_steps)
        self.train_logger.add_scalar('entropy_loss', total_entropy, self.total_steps)
        self.train_logger.add_scalar('loss', total_loss, self.total_steps)


        self.optimizer.zero_grad()
        total_loss.backward()
        grad_norm = nn.utils.clip_grad_norm_(self.network.parameters(), config.gradient_clip)
        self.train_logger.add_scalar('grad_norm', grad_norm, self.total_steps)
        self.optimizer.step()


"""
PPO_recurrent_agent
===================
"""
import numpy as np
import torch
import torch.nn as nn


from conformer_rl.agents.base_ac_agent_recurrent import BaseACAgentRecurrent
from conformer_rl.config import Config
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class PPORecurrentAgent(BaseACAgentRecurrent):
    """ Implements agent that uses the PPO (proximal policy optimization) [1]_ algorithm
    with support for recurrent neural networks.

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by this agent.

    Notes
    -----
    *Config parameters:* The following parameters are required in the `config` object. See :class:`~conformer_rl.config.agent_config.Config`
    for more details on the parameters.

    * tag
    * train_env
    * eval_env
    * optimizer_fn
    * network
    * num_workers
    * rollout_length
    * max_steps
    * save_interval
    * eval_interval
    * eval_episodes
    * recurrence
    * optimization_epochs
    * mini_batch_size
    * discount
    * use_gae
    * gae_lambda
    * entropy_weight
    * value_loss_coefficient
    * gradient_clip
    * ppo_ratio_clip
    * data_dir
    * use_tensorboard

    *Logged values*: The following values are logged during training:
    
    * advantages
    * loss
    * policy_loss
    * entropy_loss
    * value_loss
    * episodic_return_eval (total rewards per episode for eval episodes)
    * episodic_return_train (total rewards per episode for training episodes)

    References
    ----------
    .. [1] `PPO Paper <https://arxiv.org/abs/1707.06347>`_

    """
    def __init__(self, config: Config):
        super().__init__(config)
        assert config.rollout_length % self.recurrence == 0
        assert config.mini_batch_size % self.recurrence == 0

    def _train(self) -> None:
        config = self.config
        storage = self.storage

        actions = storage.order('a')
        log_probs_old = storage.order('log_pi_a')
        returns = torch.stack(self.returns, 1).view(config.num_workers * config.rollout_length, -1)
        advantages = torch.stack(self.advantages, 1).view(config.num_workers * config.rollout_length, -1)

        recurrent_states = [storage.order(f'recurrent_states_{i}') for i in range(self.num_recurrent_units)]
        states = storage.order('states')

        self.train_logger.add_scalar('advantages', advantages.mean(), self.total_steps)
        advantages = (advantages - advantages.mean()) / advantages.std()

        ############################################################################################
        #Training Loop
        ############################################################################################
        for _ in range(config.optimization_epochs):
            indices = np.arange(0, self.config.rollout_length * self.config.num_workers, self.recurrence)
            indices = np.random.permutation(indices)

            num_indices = config.mini_batch_size // self.recurrence
            starting_batch_indices = [indices[i:i+num_indices] for i in range(0, len(indices), num_indices)]
            for starting_indices in starting_batch_indices:
                batch_entropy = 0
                batch_value_loss = 0
                batch_policy_loss = 0
                batch_loss = 0

                sampled_recurrent_states = (recurrent_states[i][:, starting_indices]for i in range(self.num_recurrent_units))

                for i in range(self.recurrence):
                    sampled_actions = actions[starting_indices + i]
                    sampled_log_probs_old = log_probs_old[starting_indices + i]
                    sampled_returns = returns[starting_indices + i]
                    sampled_advantages = advantages[starting_indices + i]

                    sampled_states = [states[j] for j in (starting_indices + i)]

                    prediction, sampled_recurrent_states = self.network(sampled_states, sampled_recurrent_states, sampled_actions)


                    entropy = prediction['ent'].mean()

                    ratio = (prediction['log_pi_a'] - sampled_log_probs_old).exp()

                    obj = ratio * sampled_advantages
                    obj_clipped = ratio.clamp(1.0 - self.config.ppo_ratio_clip,
                                            1.0 + self.config.ppo_ratio_clip) * sampled_advantages

                    policy_loss = -torch.min(obj, obj_clipped).mean() - config.entropy_weight * entropy

                    value_loss = 0.5 * (sampled_returns - prediction['v']).pow(2).mean()

                    loss = policy_loss + config.value_loss_weight * value_loss

                    batch_entropy += entropy.item()
                    batch_policy_loss += policy_loss.item()
                    batch_value_loss += value_loss.item()
                    batch_loss += loss

                    if i < self.recurrence - 1:
                        for rstate_id, rstate in enumerate(recurrent_states):
                            rstate[:, starting_indices + i + 1] = sampled_recurrent_states[rstate_id].detach()


                batch_entropy /= self.recurrence
                batch_policy_loss /= self.recurrence
                batch_value_loss /= self.recurrence
                batch_loss /= self.recurrence

                self.train_logger.add_scalar('entropy_loss', batch_entropy, self.total_steps)
                self.train_logger.add_scalar('policy_loss', batch_policy_loss, self.total_steps)
                self.train_logger.add_scalar('value_loss', batch_value_loss, self.total_steps)
                self.train_logger.add_scalar('loss', batch_loss, self.total_steps)

                self.optimizer.zero_grad()
                batch_loss.backward()
                nn.utils.clip_grad_norm_(self.network.parameters(), config.gradient_clip)
                self.optimizer.step()


"""
Mol_config
==========
"""

class MolConfig:
    """Configuration object for environments.

    Specifies parameters for :class:`~conformer_rl.environments.conformer_env.ConformerEnv`
    conformer environments.

    Attributes
    ----------
    mol : rdkit Mol, required for all environments
        The molecule to be used by the environment.
    seed: int, required for all environments
        Seed for generating initial conformers for the molecule. If set to -1,
        the seed is randomized.
    
    E0: float, required for environments that use Gibbs score reward.
        The normalizing :math:`E_0` parameter for calculating Gibbs Score for the molecule. See [1]_ for more details.
    Z0 : float, required for environments that use Gibbs score reward.
        The normalizing :math:`Z_0` parameter for calculating Gibbs score for the molecule. See [1]_ for more details.
    tau : float, required for environments that use Gibbs score reward.
        The temperature (τ) parameter for calculating Gibbs Score for the molecule. See [1]_ for more details.

    pruning_thresh : float, required for environments that use pruning.
        The minimum allowed TFD (torsional fingerprint deviation) between conformers when pruning.

    References
    ----------
    .. [1] `TorsionNet paper <https://arxiv.org/abs/2006.07078>`_
    
    """

    def __init__(self):
        self.mol = None
        self.seed = -1

        # Parameters for using Gibbs Score
        self.E0 = 1
        self.Z0 = 1
        self.tau = 503

        # Parameters used for pruning 
        self.pruning_thresh = 0.05



"""
Reward_mixins
=============

Pre-built reward handlers.
"""
import numpy as np
from rdkit import Chem

from conformer_rl.utils import get_conformer_energies, get_conformer_energy, prune_conformers, prune_last_conformer

KB = 0.001985875 # Boltzmann constant in kcal/(mol * K)

# Gibbs Score
class GibbsRewardMixin:
    """Implements the Gibbs Score reward [1]_, but distance metric between conformers are judged by
    whether the conformers were produced by the same action input, instead of with TFD (Torsional Fingerprint Deviation).

    References
    ----------
    .. [1] `TorsionNet paper <https://arxiv.org/abs/2006.07078>`_
    """
    def reset(self):
        self.seen = set()
        self.repeats = 0
        self.episode_info['repeats'] = 0
        return super().reset()

    def _reward(self) -> float:
        """
        Notes
        -----
        Logged parameters:

        * energy (float): the energy of the current conformer
        * repeat (int): total number of repeated actions so far in the episode
        """
        config = self.config

        energy = get_conformer_energy(self.mol)
        self.step_info['energy'] = energy

        if tuple(self.action) in self.seen:
            self.repeats += 1
            self.episode_info['repeat'] = self.repeats
            return 0.
        else:
            self.seen.add(tuple(self.action))
            reward = np.exp(-1. * (energy - config.E0) / (KB * config.tau)) / config.Z0
            return reward

class GibbsEndPruningRewardMixin:
    """Implements the Gibbs Score reward [1]_, except overly similar conformers are only pruned at the end of an episode
    and therefore is only reflected in the final reward in each episode.

    """
    def reset(self):
        self.backup_mol = Chem.Mol(self.mol)
        self.backup_mol.RemoveAllConformers()
        return super().reset()

    def _reward(self) -> float:
        """
        Notes
        -----
        Logged parameters:

        * energy (float): the energy of the current conformer
        """
        config = self.config

        self.backup_mol.AddConformer(self.conf, assignId=True)

        energy = get_conformer_energy(self.mol)
        self.step_info['energy'] = energy

        reward = np.exp(-1. * (energy - config.E0) / (KB * config.tau)) / config.Z0

        if self._done():
            reward -= self._pruning_penalty()
        return reward

    def _pruning_penalty(self):
        config = self.config

        before_total = np.exp(-1.0 * (get_conformer_energies(self.backup_mol) - config.E0) / (KB * config.tau)).sum() / config.Z0
        self.backup_mol = prune_conformers(self.backup_mol, config.pruning_thresh)
        after_total = np.exp(-1.0 * (get_conformer_energies(self.backup_mol) - config.E0) / (KB * config.tau)).sum() / config.Z0
        return before_total - after_total

class GibbsPruningRewardMixin:
    """Implements the Gibbs Score reward [1]_.


    """
    def reset(self):
        self.backup_mol = Chem.Mol(self.mol)
        self.backup_mol.RemoveAllConformers()
        self.backup_energys = []
        return super().reset()

    def _reward(self) -> float:
        """
        Notes
        -----
        Logged parameters:

        * energy (float): the energy of the current conformer
        """
        config = self.config
        
        self.backup_mol.AddConformer(self.conf, assignId=True)
        energy = get_conformer_energy(self.mol)
        self.step_info['energy'] = energy
        self.backup_energys.append(energy)

        self._prune_conformers()
        total_reward = np.exp(-1.0 * (np.array(self.backup_energys) - config.E0) / (KB * config.tau)).sum() / config.Z0
        rew = total_reward - self.total_reward
        return rew

    def _prune_conformers(self):
        config = self.config
        self.backup_mol, self.backup_energys = prune_last_conformer(self.backup_mol, config.pruning_thresh, self.backup_energys)
        assert self.backup_mol.GetNumConformers() == len(self.backup_energys)

class GibbsLogPruningRewardMixin(GibbsPruningRewardMixin):
    """Implements the log of the Gibbs Score reward [1]_.

    """
    def _reward(self) -> float:
        """
        Notes
        -----
        Logged parameters:
        
        * energy (float): the energy of the current conformer
        """
        config = self.config

        self.backup_mol.AddConformer(self.conf, assignId=True)
        energy = get_conformer_energy(self.mol)
        self.step_info['energy'] = energy
        self.backup_energys.append(energy)

        self._prune_conformers()

        total_reward = np.log(np.exp(-1.0 * (np.array(self.backup_energys) - config.E0) / (KB * config.tau)).sum() / config.Z0)
        if not np.isfinite(total_reward):
            total_reward = np.finfo(np.float64).eps

        rew = total_reward - self.total_reward
        return rew


"""
Agent_config
============
"""

class Config:
    """Configuration object for agents.

    Specifies parameters and hyperparameters for agents. See attributes below for details.

    Attributes
    ----------
    tag : str, required for all agents
        Used to identify the training run in saved log files and in Tensorboard.
    train_env : wrapper for environments from :func:`~conformer_env.environments.environment_wrapper.Task`, required for all agents
        Wrapper for environments used to train the agents.
    eval_env : wrapper for environments from :func:`~conformer_env.environments.environment_wrapper.Task`, optional
        Wrapper for environment used to evaluate the agent.
    optimizer_fn : pytorch optimizer (torch.optim.Optimizer), required for all agents
        Optimizer function for training the neural network.
    network : pytorch neural network module (torch.nn.Module), required for all agents
        Neural network to be used by the agent.

    num_workers : int, required by all agents
        Number of parallel environments to sample from during training.
    rollout_length : int, required by all agents
        Number of environment steps taken by each worker during each sampling iteration.
    max_steps : int, required by all agents
        Number of environment steps to take before ending agent training.
    save_interval : int, required by all agents
        How often (in environment steps) to save neural network parameters. If set to 0,
        parameters will not be saved.
    eval_interval : int, required by all agents
        How often to evaluate the agent on the eval environment.
    eval_episodes : int, required by all agents
        How many episodes to evaluate the agent during each evaluation.
    recurrence : int, required by recurrent agents
        Number of steps taken before resetting recurrent states when training agent/updating network weights.
    optimization_epochs : int
        Number of epochs for training each minibatch. Used for PPO and PPORecurrent agents.
    mini_batch_size : int
        Size of each mini batch to train on. Used for PPO and PPORecurrent agents.

    discount : float, required by all agents.
        Discount factor (often denoted by γ) used for advantage estimation.
    use_gae : bool, required by all agents.
        Determines whether to use generalized advantage estimation (GAE) for estimating advantages, or
        SARSA update.
    gae_lambda : float, required by all agents if `use_gae` is ``True``
        The λ parameter used in by the generalized advantage estimator. See [1]_ for details.
    entropy_weight : float, required by all agents
        Coefficient for the entropy when calculating total loss.
    value_loss_coefficient : float, required by all agents
        Coefficient for the value loss when calculating total loss.
    gradient_clip : float, required by all agents
        Max norm for clipping gradients for neural network.
    ppo_ratio_clip : float, required by PPO and PPORecurrent agents.
        Clipping parameter ε for PPO algorithm, see [2]_ for details.

    data_dir : str, required by all agents
        Directory path for saving log files.
    use_tensorboard : bool, required by all agents
        Whether or not to save agent information to Tensorboard.



    

    References
    ----------
    .. [1] `Generalized Advantage Estimation (GAE) paper <https://arxiv.org/abs/1506.02438>`_
    .. [2] `PPO Paper <https://arxiv.org/abs/1707.06347>`_
    

    """
    def __init__(self):

        # naming
        self.tag = 'test'

        # training objects
        self.train_env = None
        self.eval_env = None
        self.optimizer_fn = None
        self.network = None
        # self.curriculum = None

        # batch hyperparameters
        self.num_workers = 1
        self.rollout_length = None
        self.max_steps = 1000000
        self.save_interval = 0
        self.eval_interval = 0
        self.eval_episodes = 2
        self.recurrence = 1
        self.optimization_epochs = 4
        self.mini_batch_size = 64

        # training hyperparameters
        self.discount = 0.9999
        self.use_gae = True
        self.gae_lambda = 0.95
        self.entropy_weight = 0.001
        self.value_loss_weight = 0.25
        self.gradient_clip = 0.5
        self.ppo_ratio_clip = 0.2

        # logging config
        self.data_dir = 'data'
        self.use_tensorboard = True
















"""
Chemistry Utilities
===================

Chemistry and molecule utility functions.
"""
import numpy as np
import bisect
from rdkit.Chem import TorsionFingerprints
import rdkit.Chem.AllChem as Chem

from typing import Tuple, List


def get_conformer_energies(mol: Chem.Mol) -> List[float]:
    """Returns a list of energies for each conformer in `mol`.
    """
    energies = []
    Chem.MMFFSanitizeMolecule(mol)
    mmff_props = Chem.MMFFGetMoleculeProperties(mol)
    for conf in mol.GetConformers():
        ff = Chem.MMFFGetMoleculeForceField(mol, mmff_props, confId=conf.GetId())
        energy = ff.CalcEnergy()
        energies.append(energy)
    
    return np.asarray(energies, dtype=float)

def get_conformer_energy(mol: Chem.Mol, confId: int = None) -> float:
    """Returns the energy of the conformer with `confId` in `mol`.
    """
    if confId is None:
        confId = mol.GetNumConformers() - 1
    Chem.MMFFSanitizeMolecule(mol)
    mmff_props = Chem.MMFFGetMoleculeProperties(mol)
    ff = Chem.MMFFGetMoleculeForceField(mol, mmff_props, confId=confId)
    energy = ff.CalcEnergy()

    return energy

def prune_last_conformer(mol: Chem.Mol, tfd_thresh: float, energies: List[float]) -> Tuple[Chem.Mol, List[float]]:
    """Prunes the last conformer of the molecule.

    If no conformers in `mol` have a TFD (Torsional Fingerprint Deviation) with the last conformer of less than `tfd_thresh`,
    the last conformer is kept. Otherwise, the lowest energy conformer with TFD less than `tfd_thresh` is kept and all other conformers
    are discarded.

    Parameters
    ----------
    mol : RDKit Mol
        The molecule to be pruned. The conformers in the molecule should be ordered by ascending energy.
    tfd_thresh : float
        The minimum threshold for TFD between conformers.
    energies : list of float
        A list of all the energies of the conformers in `mol`.

    Returns
    -------
    mol : RDKit Mol
        The updated molecule after pruning, with conformers sorted by ascending energy.
    energies : list of float
        A list of all the energies of the conformers in `mol` after pruning and sorting by ascending energy.
    """
    if tfd_thresh < 0 or mol.GetNumConformers() <= 1:
        return mol, energies

    idx = bisect.bisect(energies[:-1], energies[-1])
    tfd = TorsionFingerprints.GetTFDBetweenConformers(mol, range(0, mol.GetNumConformers() - 1), [mol.GetNumConformers() - 1], useWeights=False)
    tfd = np.array(tfd)

    # if lower energy conformer is within threshold, drop new conf
    if not np.all(tfd[:idx] >= tfd_thresh):
        energies = energies[:-1]
        mol.RemoveConformer(mol.GetNumConformers() - 1)
        return mol, energies
    else:
        keep = list(range(0, idx))
        keep.append(mol.GetNumConformers() - 1)
        keep += [x for x in range(idx, mol.GetNumConformers() - 1) if tfd[x] >= tfd_thresh]

        new = Chem.Mol(mol)
        new.RemoveAllConformers()
        for i in keep:
            conf = mol.GetConformer(i)
            new.AddConformer(conf, assignId=True)

        return new, [energies[i] for i in keep]

def prune_conformers(mol: Chem.Mol, tfd_thresh: float) -> Chem.Mol:
    """Prunes all the conformers in the molecule.

    Removes conformers that have a TFD (torsional fingerprint deviation) lower than
    `tfd_thresh` with other conformers. Lowest energy conformers are kept.

    Parameters
    ----------
    mol : RDKit Mol
        The molecule to be pruned.
    tfd_thresh : float
        The minimum threshold for TFD between conformers.

    Returns
    -------
    mol : RDKit Mol
        The updated molecule after pruning.
    """
    if tfd_thresh < 0 or mol.GetNumConformers() <= 1:
        return mol

    energies = get_conformer_energies(mol)
    tfd = tfd_matrix(mol)
    sort = np.argsort(energies)  # sort by increasing energy
    keep = []  # always keep lowest-energy conformer
    discard = []

    for i in sort:
        this_tfd = tfd[i][np.asarray(keep, dtype=int)]
        # discard conformers within the tfd threshold
        if np.all(this_tfd >= tfd_thresh):
            keep.append(i)
        else:
            discard.append(i)

    # create a new molecule to hold the chosen conformers
    # this ensures proper conformer IDs and energy-based ordering
    new = Chem.Mol(mol)
    new.RemoveAllConformers()
    for i in keep:
        conf = mol.GetConformer(int(i))
        new.AddConformer(conf, assignId=True)

    return new

def tfd_matrix(mol: Chem.Mol) -> np.array:
    """Calculates the TFD matrix for all conformers in a molecule.
    """
    tfd = TorsionFingerprints.GetTFDMatrix(mol, useWeights=False)
    n = int(np.sqrt(len(tfd)*2))+1
    idx = np.tril_indices(n, k=-1, m=n)
    matrix = np.zeros((n,n))
    matrix[idx] = tfd
    matrix += np.transpose(matrix)
    return matrix

def calculate_normalizers(mol: Chem.Mol, num_confs: int = 200, pruning_thresh: float = 0.05) -> Tuple[float, float]:
    """Calculates the :math:`E_0` and :math:`Z_0` normalizing constants for a molecule used in the TorsionNet [1]_ paper.

    Parameters
    ----------
    mol : RDKit Mol
        The molecule of interest.
    num_confs : int
        The number of conformers to generate when calculating the constants. Should equal
        the number of steps for each episode of the environment containing this molecule.
    pruning_thresh : float
        TFD threshold for pruning the conformers of `mol`.

    References
    ----------
    .. [1] `TorsionNet paper <https://arxiv.org/abs/2006.07078>`_
    """
    Chem.MMFFSanitizeMolecule(mol)
    confslist = Chem.EmbedMultipleConfs(mol, numConfs = num_confs, useRandomCoords=True)
    if (len(confslist) < 1):
        raise Exception('Unable to embed molecule with conformer using rdkit')
    Chem.MMFFOptimizeMoleculeConfs(mol, nonBondedThresh=10.)
    mol = prune_conformers(mol, pruning_thresh)
    energys = get_conformer_energies(mol)
    E0 = energys.min()
    Z0 = np.sum(np.exp(-(energys - E0)))

    mol.RemoveAllConformers()

    return E0, Z0


"""
Generate_branched_alkane
========================
"""
from rdkit import Chem
import numpy as np
import random

def generate_branched_alkane(num_atoms: int, save: bool=False) -> Chem.Mol:
    """Generates a branched alkane.

    Parameters
    ----------
    num_atoms : int
        Number of atoms in molecule to be generated.
    save : bool
        Whether to save the molecule as a .mol file.
    """
    mol = Chem.MolFromSmiles('CCCC')
    edit_mol = Chem.RWMol(mol)
    while edit_mol.GetNumAtoms() < num_atoms:
        x = Chem.rdchem.Atom(6)
        randidx = np.random.randint(len(edit_mol.GetAtoms()))
        atom = edit_mol.GetAtomWithIdx(randidx)
        if atom.GetDegree() > 2:
            continue
        if atom.GetDegree() == 2 and random.random() <= 0.5:
            continue
        idx = edit_mol.AddAtom(x)
        edit_mol.AddBond(idx, randidx, Chem.rdchem.BondType.SINGLE)

    Chem.SanitizeMol(edit_mol)
    mol = Chem.rdmolops.AddHs(edit_mol.GetMol())

    if save:
        Chem.rdmolfiles.MolToMolFile(mol, f'{num_atoms}_branched_alkane.mol')
    return mol




"""
Base_agent_recurrent
====================
"""
import torch
from conformer_rl.utils import current_time, load_model, save_model, mkdir, to_np
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


from conformer_rl.agents.base_agent import BaseAgent

from typing import Any, Tuple

class BaseAgentRecurrent(BaseAgent):
    """Base interface for building reinforcement learning agents with support for
    recurrent neural networks.

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by specific pre-built agents.

    """
    def _eval_step(self, state: object, rstates: Any = None) -> Tuple[Any, Any]:
        """Evalutes the agent on a single step of an episode of the evaluation environment.

        Parameters
        ----------
        state: object
            The current observation from the environment.
        rstates: Any
            Recurrent states from the previous iteration of the neural network.
            If none are supplied, they will be automatically initialized by the
            neural network.

        Returns
        -------
        prediction['a']: Any
            The action to be taken in the next step of the environment.
        rstates: Any
            The next recurrent states to be inputted into the neural network.

        """
        with torch.no_grad():
            prediction, rstates = self.network(state, rstates)

            return prediction['a'], rstates

    def _eval_episode(self) -> dict:
        """Evalutes the agent on a single episode of the evaluation environment.

        Returns
        -------
        dict
            Information from the evaluation environment to be logged by the
            `eval_logger`.
        """
        env = self.config.eval_env
        state = env.reset()
        done = False
        rstates = None
        info = None

        while not done:
            action, rstates = self._eval_step(state, rstates)
            state, reward, done, info = env.step(to_np(action))
            self.eval_logger.log_step(info[0]['step_info'])
        return info[0]['episode_info']



"""
Generate_xor_gate
=================
"""
from rdkit import Chem
from rdkit.Chem import Draw
import stk
from itertools import cycle, islice

def generate_xor_gate(gate_complexity: int=2, num_gates: int=3) -> Chem.Mol:
    """Generates xorgate molecule.
    """
    xorgate = XorGate(gate_complexity, num_gates)
    xorgate = xorgate.polymer.to_rdkit_mol()
    xorgate.UpdatePropertyCache()
    Chem.rdmolops.FastFindRings(xorgate)
    xorgate = Chem.AddHs(xorgate)
    return xorgate

class XorGate:
    def __init__(self, gate_complexity, num_gates):
        # use stk to construct an XOR gate molecule to design specifications
        self.gate_complexity = gate_complexity
        self.num_gates = num_gates
        self.num_torsions = num_gates * gate_complexity

        # construct XOR gate monomers
        xor_gate_top = self.make_xor_monomer(position=0)
        xor_gate_bottom = self.make_xor_monomer(position=3)
        
        # Example: for gate_complexity == 2, num_gates == 5, gives 'AABBAABBAA'
        monomer_pattern = ''.join(islice(cycle('A' * gate_complexity + 'B' * gate_complexity),
                                 num_gates * gate_complexity))
        self.polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(xor_gate_top, xor_gate_bottom),
                repeating_unit=monomer_pattern,
                num_repeating_units=1,
            )
        )

    def init_building_block(self, smiles):
        mol = stk.BuildingBlock(smiles=smiles)
        mol = Chem.rdmolops.RemoveHs(mol.to_rdkit_mol(), sanitize=True)
        # convert rdkit aromatic bonds to single and double bonds for portability
        Chem.rdmolops.Kekulize(mol)
        return stk.BuildingBlock.init_from_rdkit_mol(mol)

    def make_xor_monomer(self, position=0):
        # initialize building blocks
        benzene = self.init_building_block(smiles='C1=CC=CC=C1')
        acetaldehyde = self.init_building_block(smiles='CC=O')
        benzene = benzene.with_functional_groups([stk.SingleAtom(stk.C(position))])
        acetaldehyde = acetaldehyde.with_functional_groups(
            [stk.SingleAtom(stk.C(1))])

        # construct xor gate monomer
        xor_gate = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(benzene, acetaldehyde),
                repeating_unit='AB',
                num_repeating_units=1
            )
        )

        # construct functional groups for xor gate monomer
        # numbering starts at top and proceeds clockwise
        c_0, c_1, c_2, c_3, c_4, c_5 = stk.C(0), stk.C(
            1), stk.C(2), stk.C(3), stk.C(4), stk.C(5)
        functional_groups = [stk.GenericFunctionalGroup(atoms=(c_0, c_1, c_2, c_3, c_4, c_5),
                                                        bonders=(c_0, c_3), deleters=(c_4, c_5)),
                            stk.GenericFunctionalGroup(atoms=(c_0, c_1, c_2, c_3, c_4, c_5),
                                                        bonders=(c_1, c_2), deleters=())]
        return stk.BuildingBlock.init_from_molecule(xor_gate, functional_groups=functional_groups)



"""
Pre-built Environments
======================

This module contains several pre-built experiments. Each pre-built environment is created by overriding the following components:

* **Action Handler** refers to overriding of the :meth:`~conformer_rl.environments.conformer_env.ConformerEnv._step` method of
  :class:`~conformer_rl.environments.conformer_env.ConformerEnv`, 
  which determines how the molecule is modified given some action.
* **Reward Handler** refers to overriding of the :meth:`~conformer_rl.environments.conformer_env.ConformerEnv._reward` method of
  :class:`~conformer_rl.environments.conformer_env.ConformerEnv`, 
  which determines how the reward is calculated based on the current configuration of the molecule.
* **Observation Handler** refers to overriding of the :meth:`~conformer_rl.environments.conformer_env.ConformerEnv._obs` method of
  :class:`~conformer_rl.environments.conformer_env.ConformerEnv`, 
  which returns an observation object based on the current configuration of the molecule and is a compatible input for the neural net being used for training.

All pre-built environments inherit from :class:`~conformer_rl.environments.conformer_env.ConformerEnv` and share the same constructor.

"""

from conformer_rl.environments.conformer_env import ConformerEnv
from conformer_rl.environments.environment_components.action_mixins import ContinuousActionMixin, DiscreteActionMixin
from conformer_rl.environments.environment_components.reward_mixins import GibbsRewardMixin, GibbsPruningRewardMixin, GibbsEndPruningRewardMixin, GibbsLogPruningRewardMixin
from conformer_rl.environments.environment_components.obs_mixins import GraphObsMixin, AtomCoordsTypeGraphObsMixin



class DiscreteActionEnv(DiscreteActionMixin, GraphObsMixin, ConformerEnv):
    """
    * Action Handler: :class:`~conformer_rl.environments.environment_components.action_mixins.DiscreteActionMixin`
    * Reward Handler: default reward handler from :class:`~conformer_rl.environments.conformer_env.ConformerEnv`
    * Observation Handler: :class:`~conformer_rl.environments.environment_components.obs_mixins.GraphObsMixin`
    """
    pass


class GibbsScoreEnv(GibbsRewardMixin, DiscreteActionMixin, AtomCoordsTypeGraphObsMixin, ConformerEnv):
    """
    * Action Handler: :class:`~conformer_rl.environments.environment_components.action_mixins.DiscreteActionMixin`
    * Reward Handler: :class:`~conformer_rl.environments.environment_components.reward_mixins.GibbsRewardMixin`
    * Observation Handler: :class:`~conformer_rl.environments.environment_components.obs_mixins.AtomCoordsTypeGraphObsMixin`
    """
    pass


class GibbsScorePruningEnv(GibbsPruningRewardMixin, DiscreteActionMixin, AtomCoordsTypeGraphObsMixin, ConformerEnv):
    """
    * Action Handler: :class:`~conformer_rl.environments.environment_components.action_mixins.DiscreteActionMixin`
    * Reward Handler: :class:`~conformer_rl.environments.environment_components.reward_mixins.GibbsPruningRewardMixin`
    * Observation Handler: :class:`~conformer_rl.environments.environment_components.obs_mixins.AtomCoordsTypeGraphObsMixin`
    """
    pass


class GibbsScoreEndPruningEnv(GibbsEndPruningRewardMixin, DiscreteActionMixin, AtomCoordsTypeGraphObsMixin, ConformerEnv):
    """
    * Action Handler: :class:`~conformer_rl.environments.environment_components.action_mixins.DiscreteActionMixin`
    * Reward Handler: :class:`~conformer_rl.environments.environment_components.reward_mixins.GibbsEndPruningRewardMixin`
    * Observation Handler: :class:`~conformer_rl.environments.environment_components.obs_mixins.AtomCoordsTypeGraphObsMixin`
    """
    pass


class GibbsScoreLogPruningEnv(GibbsLogPruningRewardMixin, DiscreteActionMixin, AtomCoordsTypeGraphObsMixin, ConformerEnv):
    """
    * Action Handler: :class:`~conformer_rl.environments.environment_components.action_mixins.DiscreteActionMixin`
    * Reward Handler: :class:`~conformer_rl.environments.environment_components.reward_mixins.GibbsLogPruningRewardMixin`
    * Observation Handler: :class:`~conformer_rl.environments.environment_components.obs_mixins.AtomCoordsTypeGraphObsMixin`
    """
    pass


"""
Miscellaneous Utilities
=======================

Miscellaneous utility functions.
"""
import numpy as np
import os
import torch
from pathlib import Path

from datetime import datetime

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def current_time() -> str:
    """Returns a string containing the current date and time.
    """
    now = datetime.now()
    date_string = now.strftime("%d-%m-%Y_%H:%M:%S")
    return date_string

def set_one_thread() -> None:
    """Sets the number of CPU threads to 1 for each Python process.
    """
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['MKL_NUM_THREADS'] = '1'
    torch.set_num_threads(1)

def mkdir(path: str) -> None:
    """Creates directory specified by the input string.
    """
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)

def to_np(t: torch.tensor) -> np.array:
    """Converts a PyTorch tensor to a Numpy array.
    """
    return t.cpu().detach().numpy()

def save_model(model: torch.nn.Module, filename: str) -> None:
    """Saves model parameters of a PyTorch neural network to a file.
    """
    state_dict = model.state_dict()
    torch.save(state_dict, filename)

def load_model(model: torch.nn.Module, filename: str) -> None:
    """Loads model parameters of a PyTorch network from a file.
    """
    state_dict = torch.load(filename)
    model.load_state_dict(state_dict)



"""
Action_mixins
=============

Pre-built action handlers.
"""

from rdkit import Chem
from typing import List

class ContinuousActionMixin:
    """For each torsion of the molecule, modifies the torsion given an angle from a continuous range.
    """

    def _step(self, action: List[float]) -> None:
        """Sets the torsion angles of the molecule.

        Parameters
        ----------
        action : list of float
            Each element of `action` specifies the angle (in degrees) to set the angle of the corresponding
            torsion in the molecule.

        Notes
        -----
        Logged parameters:

        * conf: the current generated conformer is saved to the episodic mol object.
        """
        conf = self.conf
        for idx, tors in enumerate(self.nonring):
            Chem.rdMolTransforms.SetDihedralDeg(conf, *tors, float(action[idx]))
        Chem.AllChem.MMFFOptimizeMolecule(self.mol, maxIters=500, nonBondedThresh=10., confId=self.mol.GetNumConformers() - 1)
        self.episode_info['mol'].AddConformer(self.conf, assignId=True)
    
class DiscreteActionMixin:
    """For each torsion of the molecule, modifies the torsion given an angle from a discrete set of possible angles.
    """

    def _step(self, action: List[int]) -> None:
        """Sets the torsion angles of the molecule.

        Parameters
        ----------
        action : list of int between 0 and 5
            For each element of `action`, sets the corresponding torsion angle to 60 times the element degrees.
        
        Notes
        -----
        Logged parameters:
        
        * conf: the current generated conformer is saved to the episodic mol object.
        """
        for idx, tors in enumerate(self.nonring):
            ang = -180 + 60 * action[idx]
            Chem.rdMolTransforms.SetDihedralDeg(self.conf, *tors, float(ang))
        Chem.AllChem.MMFFOptimizeMolecule(self.mol, maxIters=500, nonBondedThresh=10., confId=self.mol.GetNumConformers() - 1)
        self.episode_info['mol'].AddConformer(self.conf, assignId=True)


"""
Environment_wrapper
===================
"""

import gym
import numpy as np


# documentation for SubprocVecEnv: https://github.com/DLR-RM/stable-baselines3/blob/master/stable_baselines3/common/vec_env/subproc_vec_env.py
from stable_baselines3.common.vec_env.subproc_vec_env import SubprocVecEnv
from .simple_vec_env import SimpleVecEnv

from typing import Union


def _make_env(env_id, seed, rank, **kwargs):
    def _thunk():
        np.random.seed(seed + rank)
        env = gym.make(env_id, **kwargs)
        return env

    return _thunk

def Task(name: str, concurrency: bool=False, num_envs: int=1, seed: int=np.random.randint(int(1e5)), **kwargs) -> Union[SubprocVecEnv, SimpleVecEnv]:
    """Returns a wrapper for wrapping multiple environments.

    Parameters
    ----------
    name : str
        The name of the environment, as registered using the ``gym.register`` method.
    concurrency : bool
        Whether or not the environments should be run in parallel across multiple CPU's.
    num_envs : bool
        The number of environments to be wrapped.
    seed : bool
        Seed for initializing the environments.

    Returns
    -------
    A wrapper for the environment(s).
    """
    envs = [_make_env(name, seed, i, **kwargs) for i in range(num_envs)]
    return SubprocVecEnv(envs) if concurrency else SimpleVecEnv(envs)


"""
Molecule_features
=================
Helper functions for extracting features from molecules conformers
to generate graph representations in :mod:`conformer_rl.environments.environment_components.obs_mixins`.
"""
from rdkit import Chem
from typing import List

def bond_type(bond: Chem.Bond) -> List[bool]:
    """Extracts features from a bond in a molecule.

    Parameters
    ----------
    bond : rdkit Bond object
        The bond to extract features from

    Returns
    -------
    List of bools of length 6
    Each element corresponds to, respectively, whether the bond is
    
        * A single bond
        * A double bond
        * A triple bond
        * An aromatic bond
        * Is conjugated
        * Is in a ring structure

    """

    bt = bond.GetBondType()
    bond_feats = []
    bond_feats = bond_feats + [
        bt == Chem.rdchem.BondType.SINGLE, bt == Chem.rdchem.BondType.DOUBLE,
        bt == Chem.rdchem.BondType.TRIPLE, bt == Chem.rdchem.BondType.AROMATIC,
        bond.GetIsConjugated(),
        bond.IsInRing()
    ]
    return bond_feats

def get_bond_pairs(mol: Chem.Mol) -> List[List[int]]:
    """Returns a list of all pairs of atoms that have a bond between each other.
    """

    bonds = mol.GetBonds()
    res = [[],[]]
    for bond in bonds:
        res[0] += [bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]
        res[1] += [bond.GetEndAtomIdx(), bond.GetBeginAtomIdx()]
    return res

def atom_coords(atom: Chem.Atom, conf: Chem.Conformer) -> List[float]:
    """Returns the x, y, and z coordinates of an atom in a molecule conformer.
    """
    p = conf.GetAtomPosition(atom.GetIdx())
    fts = [p.x, p.y, p.z]
    return fts

def atom_type_CO(atom: Chem.Atom) -> List[bool]:
    """Returns a one-hot list of length 2 for whether `atom` is a carbon or oxygen atom.
    """
    anum = atom.GetSymbol()
    atom_feats = [
        anum == 'C', anum == 'O',
    ]
    return atom_feats


"""
Graph_components
================
Modularized graph neural network components using PyTorch Geometric.
"""
from torch import nn
import torch.nn.functional as F
import torch
import torch_geometric.nn as gnn
from torch_geometric.data import Batch


device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class MPNN(torch.nn.Module):
    """
    Implements a basic unit of the message passing neural network (MPNN) [1]_.

    Parameters
    ----------
    hidden_dim : int
        Dimension of the hidden layer.
    edge_dim : int
        Dimension of the edge embeddings in the input graph.
    node_dim : int
        Dimension of the node embeddings in the input graph.
    message_passing_steps : int 
        Number of message passing steps to execute. See [1]_ for more details.

    References
    ----------
    .. [1] `MPNN paper <https://arxiv.org/abs/1704.01212>`_
    """
    def __init__(self, hidden_dim: int, edge_dim: int, node_dim: int, message_passing_steps: int=6):
        super().__init__()
        self.mpnn_iters = message_passing_steps
        self.fc = torch.nn.Linear(node_dim, hidden_dim)
        func_ag = nn.Sequential(nn.Linear(edge_dim, hidden_dim), nn.ReLU(inplace=False), nn.Linear(hidden_dim, hidden_dim * hidden_dim))
        self.conv = gnn.NNConv(hidden_dim, hidden_dim, func_ag, aggr='mean')
        self.gru = nn.GRU(hidden_dim, hidden_dim)

    def forward(self, data: Batch) -> torch.Tensor:
        out = F.relu(self.fc(data.x))
        h = out.unsqueeze(0)

        for i in range(self.mpnn_iters):
            m = F.relu(self.conv(out, data.edge_index, data.edge_attr))
            out, h = self.gru(m.unsqueeze(0), h)
            out = out.squeeze(0)
        return out

class GAT(torch.nn.Module):
    """
    Implements a basic unit of the graph attention network (GAT) [2]_.

    Parameters
    ----------
    hidden_dim : int
        Dimension of the hidden layer.
    node_dim : int
        Dimension of the node embeddings in the input graph.
    num_layers : int 
        Number of GAT conv layers. See [2]_ for more details.

    References
    ----------
    .. [2] `GAT paper <https://arxiv.org/abs/1710.10903>`_
    """
    def __init__(self, hidden_dim: int, node_dim: int, num_layers: int=6):
        super().__init__()
        self.fc = torch.nn.Linear(node_dim, hidden_dim)
        self.conv_layers = nn.ModuleList()
        self.conv_layers.append(gnn.GATConv(hidden_dim, hidden_dim, heads=2))
        for i in range(num_layers - 2):
            self.conv_layers.append(gnn.GATConv(hidden_dim*2, hidden_dim, heads=2))
        self.conv_layers.append(gnn.GATConv(hidden_dim*2, hidden_dim))

    def forward(self, data : Batch) -> torch.Tensor:
        out = F.relu(self.fc(data.x))
        for layer in self.conv_layers:
            out = layer(out, data.edge_index)
        return out


"""
Env_logger
==========
"""

import pickle
from rdkit import Chem
from conformer_rl.utils import mkdir
from typing import Any

class EnvLogger:
    """Used by the agent for logging metrics produced by the environment, for example, observations, rewards, renders, etc.
    Supports saving data to pickle and saving molecules as .mol files.

    Parameters
    ----------
    tag : str
        Unique tag for identifying the logging session.
    dir : str
        Path to root directory for where logging results should be saved.


    Attributes
    ----------
    step_data : dict
        Used for storing data for every step of a single episode. 
        Maps from strings to lists, where each index of the list corresponds
        to the data for that corresponding step within the episode.
    episode_data : dict 
        Used for storing information for a single episode. Used to store the self.step_data
        for that episode and metadata global to the entire episode.
    cache : dict 
        Used for storing data across several episodes. Maps from strings to lists, where each
        index of the list corresponds to an episode.
    """
    def __init__(self, tag: str, dir: str = "data"):
        self.dir = dir
        mkdir(dir)
        self.tag = tag
        self.step_data = {}
        self.episode_data = {}
        self.cache = {}

    def clear_data(self) -> None:
        """Resets the logger.
        """
        self.cache = {}
        self.episode_data = {}
        self.step_data = {}

    def clear_episode(self) -> None:
        """Clears episode and step data.
        """
        self.step_data = {}
        self.episode_data = {}

    def log_step_item(self, key: str, val: Any) -> None:
        """Logs a single key value pair for current step.

        If an existing key is found, the value is appended
        to the list associated with that key.

        Parameters
        ----------
        key : str
            the key for the data to be added.
        val : Any
            the value of the data to be added.
        """
        if key in self.step_data:
            self.step_data[key].append(val)
        else:
            self.step_data[key] = [val]

    def log_step(self, step_data: dict) -> None:
        """Logs each key-value pair for current step.

        If an existing key is found, the value is appended
        to the list associated with that key.

        Parameters
        ----------
        step_data : dict
            Contains key-value pairs to be logged.
        """
        for key, val in step_data.items():
            self.log_step_item(key, val)

    def log_episode_item(self, key: str, value: Any) -> None:
        """Logs a single key-value pair to the per-episode data.
        """
        self.episode_data[key] = value

    def log_episode(self, episode_data: dict) -> None:
        """Logs each key-value pair to the per-episode data.

        Also adds `step_data` to the per-episode data with corresponding key
        'step_data'. Existing keys will be overwritten.

        Parameters
        ----------
        episode_data : dict
            Contains key-value pairs to be logged.
        """
        self.episode_data.update(episode_data)
        self.episode_data["step_data"] = self.step_data

    def save_episode(self, subdir: str, save_pickle: bool = True, save_molecules: bool = False, save_cache: bool = False) -> None:
        """Saves current episode_data with options for dumping to pickle file,
        saving data to a cache dict, and saving the rdkit molecules as .mol files.
        Clears the current episode and step data.

        Parameters
        ----------
        subdir : str
            The directory for episode data to be saved (relative to self.dir)
        save_pickle : bool
            If True, dumps episode_data as a .pickle file.
        save_molecules : bool
            If True, and 'molecule' key exists in step_data, dumps each molecule generated 
            throughout the episode as a .mol file uniquely named by the step number.
        save_cache : bool
            If True, episode data is cached to self.data.
        """
        path = self.dir + '/' +  'env_data' + '/' + self.tag + '/' + subdir
        mkdir(path)
        filename = path + '/' +  'data.pickle'

        if save_pickle:
            outfile = open(filename, 'w+b')
            pickle.dump(self.episode_data, outfile)
            outfile.close()

        if save_molecules and 'mol' in self.episode_data:
            mol = self.episode_data['mol']
            for i in range(mol.GetNumConformers()):
                Chem.MolToMolFile(mol, filename=path + '/' +  f'step_{i}.mol', confId=i)

        if save_cache:
            self._add_to_cache(self.episode_data)

        self.clear_episode()

    def _add_to_cache(self, data:dict) -> None:
        """Logs each key-value pair in data to self.cache.
        If an existing key is found, the value is appended
        to the list associated with that key.

        Parameters
        ----------
        data : dict
            contains key-value pairs to be logged.
        """
        for key, val in data.items():    
            if key in self.cache:
                self.cache[key].append(val)
            else:
                self.cache[key] = [val]


"""
Base_ac_agent
=============
"""
import torch
import numpy as np
import logging
import time
from conformer_rl.utils import current_time, load_model, save_model, mkdir, to_np
from conformer_rl.config import Config
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

from conformer_rl.agents.base_agent import BaseAgent

class BaseACAgent(BaseAgent):
    """Base interface for building reinforcement learning agents that use actor-critic algorithms.

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by specific pre-built agents.

    """
    def __init__(self, config: Config):
        super().__init__(config)

        self.network = config.network # neural network / model
        self.optimizer = config.optimizer_fn(self.network.parameters())

        self.total_rewards = np.zeros(config.num_workers)
        self.states = self.task.reset()
        self.prediction = None

    def step(self) -> None:
        """Performs one iteration of acquiring samples on the environment
        and then trains on the acquired samples.
        """
        self.storage.reset()
        sample_start = time.time()
        self._sample()
        logging.debug(f'sample time: {time.time() - sample_start} seconds')
        train_start = time.time()
        self._calculate_advantages()
        self._train()
        logging.debug(f'train time: {time.time() - train_start} seconds')

    def _sample(self) -> None:
        """Collects samples from the training environment.
        """
        config = self.config
        states = self.states
        storage = self.storage
        ##############################################################################################
        #Sampling Loop
        ##############################################################################################
        for _ in range(config.rollout_length):
            self.total_steps += config.num_workers

            #run the neural net once to get prediction
            prediction = self.network(states)

            #step the environment with the action determined by the prediction
            next_states, rewards, terminals, _ = self.task.step(to_np(prediction['a']))
            self.total_rewards += np.asarray(rewards)

            for idx, done in enumerate(terminals):
                if done:
                    logging.info(f'logging episodic return train... {self.total_steps}')
                    self.train_logger.add_scalar('episodic_return_train', self.total_rewards[idx], self.total_steps)
                    self.total_rewards[idx] = 0.

            #add everything to storage
            storage.append(prediction)
            storage.append({
                'states': states,
                'r': torch.tensor(rewards).unsqueeze(-1).to(device),
                'm': torch.tensor(1 - terminals).unsqueeze(-1).to(device)
                })
            states = next_states


        self.states = states

        prediction = self.network(states)
        self.prediction = prediction

        storage.append(prediction)

    def _train(self) -> None:
        raise NotImplementedError

    def _calculate_advantages(self) -> None:
        """Performs advantage estimation.

        Uses either SARSA or generalized advantage estimation (GAE) for estimating advantages,
        depending on the config.
        """
        config = self.config
        storage = self.storage

        self.advantages, self.returns = [None] * config.rollout_length, [None] * config.rollout_length
        adv = torch.zeros((config.num_workers, 1), dtype=torch.float64).to(device)
        ret = self.prediction['v'].squeeze(0)

        for i in reversed(range(config.rollout_length)):
            ret = storage['r'][i] + config.discount * storage['m'][i] * ret
            if not config.use_gae:
                adv = ret - storage['v'][i].detach()
            else:
                td_error = storage['r'][i] + config.discount * storage['m'][i] * storage['v'][i + 1] - storage['v'][i]
                adv = adv * config.gae_lambda * config.discount * storage['m'][i] + td_error
            self.advantages[i] = adv.detach()
            self.returns[i] = ret.detach()


"""
Base_ac_agent_recurrent
=======================
"""
import torch
import numpy as np
from conformer_rl.utils import current_time, load_model, save_model, mkdir, to_np
from conformer_rl.config import Config
import logging
import time
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

from conformer_rl.agents.base_ac_agent import BaseACAgent
from conformer_rl.agents.base_agent_recurrent import BaseAgentRecurrent

class BaseACAgentRecurrent(BaseACAgent, BaseAgentRecurrent):
    """Base interface for building reinforcement learning agents that use actor-critic algorithms
    with support for recurrent neural networks"

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by specific pre-built agents.

    """
    def __init__(self, config: Config):
        super().__init__(config)

        with torch.no_grad():
            _, self.recurrent_states = self.network(self.states)
            self.num_recurrent_units = len(self.recurrent_states)
        for state in self.recurrent_states:
            state.zero_()

        self.recurrence = self.config.recurrence

    def step(self) -> None:
        """Performs one iteration of acquiring samples on the environment
        and then trains on the acquired samples.
        """
        self.storage.reset()
        with torch.no_grad():
            sample_start = time.time()
            self._sample()
            logging.debug(f'sample time: {time.time() - sample_start} seconds')
        train_start = time.time()
        self._calculate_advantages()
        self._train()
        logging.debug(f'train time: {time.time() - train_start} seconds')
        
    def _sample(self) -> None:
        """Collects samples from the training environment.
        """
        config = self.config
        states = self.states
        storage = self.storage
        ##############################################################################################
        #Sampling Loop
        ##############################################################################################
        for _ in range(config.rollout_length):
            self.total_steps += config.num_workers

            #add recurrent states (lstm hidden and lstm cell states) to storage
            storage.append({f'recurrent_states_{i}' : rstate for i, rstate in enumerate(self.recurrent_states)})

            #run the neural net once to get prediction
            prediction, self.recurrent_states = self.network(states, self.recurrent_states)

            #step the environment with the action determined by the prediction
            next_states, rewards, terminals, _ = self.task.step(to_np(prediction['a']))

            self.total_rewards += np.asarray(rewards)

            for idx, done in enumerate(terminals):
                if done:
                    logging.info(f'logging episodic return train... {self.total_steps}')
                    self.train_logger.add_scalar('episodic_return_train', self.total_rewards[idx], self.total_steps)
                    self.total_rewards[idx] = 0.

                    # zero out lstm states for finished environments
                    for rstate in self.recurrent_states:
                        rstate[:, idx].zero_()

            #add everything to storage
            storage.append(prediction)
            storage.append({
                'states': states,
                'r': torch.tensor(rewards).unsqueeze(-1).to(device),
                'm': torch.tensor(1 - terminals).unsqueeze(-1).to(device)
                })
            states = next_states


        self.states = states

        prediction, _ = self.network(states, self.recurrent_states)
        self.prediction = prediction

        storage.append(prediction)


"""
Conformer_env
=============
"""
import numpy as np
import gym
import copy

from rdkit.Chem import AllChem as Chem
from rdkit.Chem import TorsionFingerprints
from conformer_rl.utils import get_conformer_energy
from conformer_rl.config import MolConfig

import logging
from typing import List, Tuple, Mapping, Any

class ConformerEnv(gym.Env):
    """Base interface for building conformer generation environments.

    Parameters
    ----------
    mol_config : :class:`~conformer_rl.config.mol_config.MolConfig`
        Configuration object specifying molecule and parameters to be used in the environment.
    max_steps : int
        The number of steps before the end of an episode.

    Attributes
    ----------
    config : :class:`~conformer_rl.config.mol_config.MolConfig`
        Configuration object specifying molecule and parameters to be used in the environment.
    total_reward : float
        Keeps track of the total reward for the current episode.
    current_step : int
        Keeps track of the number of elapsed steps in the current episode.
    step_info : dict from str to list
        Used for keeping track of data obtained at each step of an episode for logging.
    episode_info : dict from str to Any
        Used for keeping track of data useful at the end of an episode, such as total_reward, for logging.
    

    """
    metadata = {'render.modes': ['human']}

    def __init__(self, mol_config: MolConfig, max_steps = 200):
        super(ConformerEnv, self).__init__()
        logging.info('initializing conformer environment')
        self.config = copy.deepcopy(mol_config)
        self.max_steps = max_steps
        self.total_reward = 0
        self.current_step = 0

        self.step_info = {}
        self.episode_info = {}

        self.mol = self.config.mol

        # set mol to have exactly one conformer
        if self.mol.GetNumConformers() != 1:
            logging.debug("Input molecule to environment should have exactly one conformer, none or more than one detected.")
            self.mol.RemoveAllConformers()
            if Chem.EmbedMolecule(self.mol, randomSeed=self.config.seed, useRandomCoords=True) == -1:
                raise Exception('Unable to embed molecule with conformer using rdkit')
        self.conf = self.mol.GetConformer()
        nonring, ring = TorsionFingerprints.CalculateTorsionLists(self.mol)
        self.nonring = [list(atoms[0]) for atoms, ang in nonring]

        self.reset()

    def step(self, action: Any) -> Tuple[object, float, bool, dict]:
        """Simulates one iteration of the environment.

        Updates the environment with the input action, and calculates the current observation,
        reward, done, and info.

        Parameters
        ----------
        action : Any, depending on implementation of :meth:`~ConformerEnv._step()`
            The action to be taken by the environment.

        Returns
        -------
        obs : Any, depending on implementation of :meth:`~ConformerEnv._obs()`
            An object reflecting the current configuration of the environment/molecule.
        reward : float
            The reward calculated given the current configuration of the environment.
        done : bool
            Whether or not the current episode has finished.
        info : dict
            Information about the current step and episode of the environment, to be used for logging.

        Notes
        -----
        Logged parameters:

        * reward (float): the reward for the current step
        """
        self.action = action

        self._step(action)
        self.current_step += 1

        assert(self.mol.GetNumConformers() == 1)

        obs = self._obs()
        reward = self._reward()
        self.step_info['reward'] = reward
        self.total_reward += reward
        done = self._done()
        info = copy.deepcopy(self._info())

        logging.info(f"step {self.current_step} reward {reward}")


        return obs, reward, done, info
        
    def reset(self) -> object:
        """Resets the environment and returns the observation of the environment.
        """
        logging.info("reset called")
        # reset environment state variables
        self.total_reward = 0
        self.current_step = 0

        self.step_info = {}
        self.episode_info['mol'] = Chem.Mol(self.mol)
        self.episode_info['mol'].RemoveAllConformers()

        obs = self._obs()

        return obs

    def _step(self, action: Any) -> None:
        """Does not modify molecule.

        Notes
        -----
        Logged parameters:

        * conf: the current generated conformer is saved to the episodic mol object.
        """
        self.episode_info['mol'].AddConformer(self.conf, assignId=True)

    def _obs(self) -> Chem.rdchem.Mol:
        """Returns the current molecule.
        """
        return self.mol

    def _reward(self) -> float:
        """Returns :math:`e^{-1 * energy}` where :math:`energy` is the
        energy of the current conformer of the molecule.

        Notes
        -----
        Logged parameters:

        * energy (float): the energy of the current conformer
        """
        energy = get_conformer_energy(self.mol)
        reward =  np.exp(-1. * energy)

        self.step_info['energy'] = energy
        return reward

    def _done(self) -> bool:
        """Returns true if the current number of elapsed steps has exceeded
        the max number of steps per episode.
        """
        return self.current_step >= self.max_steps

    def _info(self) -> Mapping[str, Mapping[str, Any]]:
        """Returns a dict wrapping `episode_info` and `step_info`.

        Notes
        -----
        Logged parameters:
        
        * total_reward (float): total reward of the episode is updated
        """
        info = {}

        if self._done():
            self.episode_info["total_rewards"] = self.total_reward
            info['episode_info'] = self.episode_info

        info['step_info'] = self.step_info

        return info

        


"""
Analysis
========

Functions for analyzing and visualizing (in Jupyter/IPython notebook) logged
environment data. The functions for visualizations here provide a basic set of functionality
to guide users in understanding the format of the logged environment data. 
Users are encouraged to generate their own plots and visualizations based on their specific needs.
"""

import pickle
import numpy as np
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt
from rdkit import Chem
from conformer_rl.utils import tfd_matrix
import py3Dmol
import logging

from typing import Any, List, Optional, Tuple

def _load_from_pickle(filename: str) -> Any:
    """Loads an object from a .pickle file.
    """
    infile = open(filename, 'rb')
    data = pickle.load(infile)
    infile.close()
    return data

def load_data_from_pickle(paths: List[str], indices: Optional[List[str]]=None) -> dict:
    """Loads saved pickled environment data from multiple runs into a combined data dict.

    Parameters
    ----------
    paths : list of str
        List of paths to .pickle files corresponding to the environment data from the runs of interest.
    indices : list of str, optional
        Specifies custom indices/labels to be displayed in generated Seaborn graphs for each run. Should be the
        same length as `paths`. If not specified, the labels default to ``test0, test1, test2, ...``.

    Returns
    -------
    dict mapping from str to list
        The str corresponds to the key for the data in the original pickled dict object. The list contains
        the data for each of the environment data sets specified in `paths`, in the same order they were
        given in `paths`.

    Notes
    -----
    The ``.pickle`` files specified by `paths` should be dumped directly by :class:`~conformer_rl.logging.env_logger.EnvLogger`,
    and should correspond to a single evaluation episode.
    See :meth:`conformer_rl.logging.env_logger.EnvLogger.save_episode` for more details on the dumped format.

    An example of how the function operates: Suppose that our paths are::

        ['data1.pickle', 'data2.pickle', 'data3.pickle']

    And each pickle object contains corresponding data::

        data1 = {
            'total_rewards': data1_total_rewards,
            'mol': data1_molecule,
            'rewards': [data1_step1_rewards, data1_step2_rewards, data1_step3_rewards, data1_step4_rewards]
        }
        data2 = {
            'total_rewards': data2_total_rewards,
            'mol': data2_molecule,
            'rewards': [data2_step1_rewards, data2_step2_rewards, data2_step3_rewards, data2_step4_rewards]
        }
        data3 = {
            'total_rewards': data3_total_rewards,
            'mol': data3_molecule,
            'rewards': [data3_step1_rewards, data3_step2_rewards, data3_step3_rewards, data3_step4_rewards]
        }
    
    Suppose that data1 corresponds to some eval data obtained from training with the PPO agent, data2 was obtained
    from the PPORecurrent agent, and data3 was obtained from training with the A2C agent. Then we can input custom `indices`
    to help us understand each dataset better::

        indices = ['PPO', 'PPO_recurrent', 'A2C']

    Given these data and indices, :func:`load_data_from_pickle` would return the following dict::

        {
            'indices': ['PPO', 'PPO_recurrent', 'A2C'],
            'total_rewards': [
                data1_total_rewards,
                data2_total_rewards,
                data3_total_rewards
            ],
            'mol': [
                data1_molecule,
                data2_molecule,
                data3_molecule
            ],
            'rewards': [
                [data1_step1_rewards, data1_step2_rewards, data1_step3_rewards, data1_step4_rewards],
                [data2_step1_rewards, data2_step2_rewards, data2_step3_rewards, data2_step4_rewards],
                [data3_step1_rewards, data3_step2_rewards, data3_step3_rewards, data3_step4_rewards]
            ],
        }
    
    This format consolidates all the data into a single dict and is compatible with the other visualization
    functions in this module. Furthermore, it is also easy to convert a dict of this format into a Pandas dataframe
    or other tabular formats if needed.
    """

    if not isinstance(paths, list):
        paths = [paths]

    if indices is None:
        indices = [f'test{i}' for i, x in enumerate(paths)]

    data = map(_load_from_pickle, paths)
    data =  list(data)

    final_data = {"indices": indices}
    for datum in data:
        if 'step_data' in datum:
            datum.update(datum['step_data'])
            del datum['step_data']
        for key, val in datum.items():
            final_data.setdefault(key, []).append(val)
            
    return final_data

def list_keys(data: dict) -> List[str]:
    """Return a list of all keys in a dict.

    Parameters
    ----------
    data : dict
        The dictionary to retrieve keys from.
    """
    return list(key for key, val in data.items())

def bar_plot_episodic(key: str, data: dict) -> matplotlib.axes.Axes:
    """Plots a bar plot comparing a scalar value across all episodes loaded in `data`.

    Parameters
    ----------
    key : str
        The key for the values to be compared across all data sets/episodes.
    data : dict
        Data dictionary generated by :meth:`load_data_from_pickle`.
    """
    ax = sns.barplot(x=data["indices"], y=data[key])
    ax.set(xlabel='run', ylabel=key)
    return ax

def histogram_select_episodes(key: str, data: dict, episodes: List[int]=None, binwidth: float=10, figsize: Tuple[float, float]=(8., 6.)) -> matplotlib.axes.Axes:
    """Plots a single histogram where data for each episode in `episodes` are overlayed.

    Parameters
    ----------
    key : str
        The key for the values to be compared across all data sets/episodes.
    data : dict
        Data dictionary generated by :meth:`load_data_from_pickle`.
    episodes : list of int, optional
        Specifies the indices in `data` for the episodes to be shown. If not specified,
        all episodes are shown.
    binwidth: float
        The width of each bin in the histogram.
    figsize: 2-tuple of float
        Specifies the size of the plot.
    """
    if episodes is None:
        episodes = list(range(len(data['indices'])))
    fig, axes = plt.subplots()
    input_data = {data["indices"][i]: data[key][i] for i in episodes}
    sns.histplot(data=input_data, binwidth=binwidth, ax=axes)
    axes.set(xlabel=key)

    return fig, axes

def histogram_episodic(key: str, data: dict, binwidth: int=10, figsize: Tuple[float, float]=(8., 6.)) -> matplotlib.axes.Axes:
    """Plots histogram on separate axes for each of the episode data sets in `data`.

    Parameters
    ----------
    key : str
        The key for the values to be compared across all data sets/episodes.
    data : dict
        Data dictionary generated by :meth:`load_data_from_pickle`.
    binwidth: float
        The width of each bin in the histogram.
    figsize: 2-tuple of float
        Specifies the size of the plot.
    """
    n = len(data["indices"])
    fig, axes = plt.subplots(nrows=n, figsize=figsize)

    if n == 1:
        sns.histplot(data={data['indices'][0]: data[key][0]}, binwidth=binwidth, ax=axes)
        axes.set(xlabel=key)
    else:
        for i, index in enumerate(data["indices"]):
            sns.histplot(data={index: data[key][i]}, binwidth=binwidth, ax=axes[i])
            axes[i].set(xlabel=key)

    return fig, axes

def heatmap_episodic(key: str, data: dict, figsize: Tuple[float, float]=(8., 6.)) -> matplotlib.axes.Axes:
    """Plots heatmap(s) for matrix data corresponding to `key` across all episodes
    loaded in `data`.

    Parameters
    ----------
    key : str
        The key for the values to be compared across all data sets/episodes.
    data : dict
        Data dictionary generated by :meth:`load_data_from_pickle`.
    figsize: 2-tuple of float
        Specifies the size of the plot.
    """
    n = len(data["indices"])
    fig, axes = plt.subplots(nrows=n, figsize=figsize)
    if n == 1:
        sns.heatmap(data=data[key][0], ax=axes)
        axes.set(xlabel=data['indices'][0])
    else:
        for i, index in enumerate(data["indices"]):
            sns.heatmap(data=data[key][i], ax=axes[i])
            axes[i].set(xlabel=index)
    
    return fig, axes

def calculate_tfd(data: str) -> None:
    """Updates data with the TFD (Torsion Fingerprint Deviation) matrix (with key 'tfd_matrix') and sum of the TFD matrix
    (with key 'tfd_total') for the molecule conformers across each episode loaded in `data`.

    Parameters
    ----------
    data : dict
        Data dictionary generated by :meth:`load_data_from_pickle`.
    """
    if not 'mol' in data:
        raise Exception('data dict must contain RDKit Mol object with \'mol\' key to generate tfd matrix.')
    if 'tfd_matrix' in data or 'tfd_total' in data:
        logging.info("tfd matrix already exists, recalculating...")
        data.pop('tfd_matrix')
        data.pop('tfd_total')
    for mol in data['mol']:
        matrix = tfd_matrix(mol)
        data.setdefault('tfd_matrix', []).append(matrix)
        data.setdefault('tfd_total', []).append(np.sum(matrix))

def drawConformer(mol: Chem.Mol, confId: int=-1, size: Tuple[int, int]=(300, 300), style: str="stick") -> py3Dmol.view:
    """Displays interactive 3-dimensional representation of specified conformer.

    Parameters
    ----------
    mol : RDKit Mol object
        The molecule containing the conformer to be displayed.
    confId : int
        The ID of the conformer to be displayed.
    size : Tuple[int, int]
        The size of the display (width, height).
    style: str
        The drawing style for displaying the molecule. Can be sphere, stick, line, cross, cartoon, and surface.
    """
    block = Chem.MolToMolBlock(mol, confId=confId)
    view = py3Dmol.view(width=size[0], height=size[1])
    view.addModel(block, 'mol')
    view.setStyle({style : {}})
    view.zoomTo()
    return view

def drawConformer_episodic(data: dict, confIds: List[int], size: Tuple[int, int]=(300, 300), style: str="stick") -> py3Dmol.view:
    """Displays a specified conformer for each episode loaded in `data`.

    Parameters
    ----------
    data : dict from string to list
        Contains the loaded episode information. 'mol' must be a key in data and the corresponding list must contain
        RDKit Mol objects.
    confIds : list of int
        The indices for the conformers to be displayed (for each episode loaded in data).
    size : Tuple[int, int]
        The size of the display for each individual molecule (width, height).
    style: str
        The drawing style for displaying the molecule. Can be sphere, stick, line, cross, cartoon, and surface.
    """
    n = len(data['mol'])
    view = py3Dmol.view(width=size[0]*n, height=size[0]*n, linked=False, viewergrid=(n, 1))
    for i in range(n):
        block = Chem.MolToMolBlock(data['mol'][i], confId=confIds[i])
        view.addModel(block, 'mol', viewer=(i, 0))
        view.setStyle({style:{}}, viewer=(i, 0))
    view.zoomTo()
    return view


"""
generate_lignin
===============
"""
import numpy as np

# Chemical Drawing
from rdkit.Chem import MolFromMolBlock
from rdkit import Chem

# Lignin-KMC functions and global variables used in this notebook
from ligninkmc.kmc_functions import run_kmc, generate_mol
from ligninkmc.create_lignin import (calc_rates, create_initial_monomers, create_initial_events,
                                     create_initial_state, analyze_adj_matrix, adj_analysis_to_stdout)
from ligninkmc.kmc_common import (DEF_E_BARRIER_KCAL_MOL, ADJ_MATRIX, MONO_LIST, MONOMER, OX, GROW, Monomer, Event)

# Calculate the rates of reaction in 1/s (or 1/monomer-s if biomolecular) at the specified temp
temp = 298.15  # K
rxn_rates = calc_rates(temp, ea_kcal_mol_dict=DEF_E_BARRIER_KCAL_MOL)

def generate_lignin(num_monomers: int = 1) -> Chem.Mol:
        """Generates lignin molecule.

        parameters
        ----------
        num_monomers : int
                Number of monomers in lignin molecule.
        """
        # Set the percentage of S
        sg_ratio = 0
        pct_s = sg_ratio / (1 + sg_ratio)

        # Set the initial and maximum number of monomers to be modeled.
        ini_num_monos = 1
        max_num_monos = num_monomers

        # Maximum time to simulate, in seconds
        t_max = 1  # seconds
        mono_add_rate = 1e4  # monomers/second

        # Use a random number and the given sg_ratio to determine the monolignol types to be initially modeled
        monomer_draw = np.random.rand(ini_num_monos)
        initial_monomers = create_initial_monomers(pct_s, monomer_draw)

        # Initially allow only oxidation events. After they are used to determine the initial state, add 
        #     GROW to the events, which allows additional monomers to be added to the reaction at the 
        #     specified rate and with the specified ratio
        initial_events = create_initial_events(initial_monomers, rxn_rates)
        initial_state = create_initial_state(initial_events, initial_monomers)
        initial_events.append(Event(GROW, [], rate=mono_add_rate))

        # simulate lignin creation
        result = run_kmc(rxn_rates, initial_state,initial_events, n_max=max_num_monos, t_max=t_max, sg_ratio=sg_ratio)
        # using RDKit
        nodes = result[MONO_LIST]
        adj = result[ADJ_MATRIX]
        block = generate_mol(adj, nodes)
        mol = MolFromMolBlock(block)
        mol = Chem.AddHs(mol)

        return mol



import numpy as np

class SimpleVecEnv():

    def __init__(self, env_fns):
        self.envs = [fn() for fn in env_fns]
        self.num_envs = len(env_fns)
        self.actions = None

    def step(self, actions):
        data = []
        for i in range(self.num_envs):
            obs, rew, done, info = self.envs[i].step(actions[i])
            if done:
                obs = self.envs[i].reset()
            data.append([obs, rew, done, info])
        obs, rew, done, info = zip(*data)
        return obs, np.asarray(rew), np.asarray(done), info

    def reset(self):
        return [env.reset() for env in self.envs]

    def close(self):
        for env in self.envs:
            env.close()

    def render(self):
        return [env.render() for env in self.envs]

    def env_method(self, method_name, *method_args, **method_kwargs):
        return [getattr(env, method_name)(*method_args, **method_kwargs) for env in self.envs]



"""
Storage
=======

Storage utility class for agents.

"""

import torch

class Storage:
    """Saves and stores experiences from the agent.

    Parameters
    ----------
    rollout : int
        The rollout length of the agent in each sampling iteration (per worker).
    workers : int
        The number of workers used for sampling by the agent.

    """
    def __init__(self, rollout: int, workers: int):
        self.storage = {}
        self.rollout = rollout
        self.workers = workers

    def __getitem__(self, key: str) -> list:
        """Returns all batched items in storage with the given `key`.

        :meta public:

        Parameters
        ----------
        key : str
            The key of the items to be accessed.

        Returns
        -------
        list
            A list of all the items with the associated `key` 
            previously appended to the storage.


        """
        return self.storage[key]

    def append(self, data:dict) -> None:
        """Appends item(s) to the storage

        Parameters
        ----------
        data : dict
            The items to be appended.
        """
        for key, val in data.items():
            self.storage.setdefault(key, []).append(val)

    def order(self, key:str) -> list:
        """Splits each batch of items associated with `key` by worker, and then
        orders the items into a list sorted firstly by worker and secondly by the
        order the items were appended to storage.

        Parameters
        ----------
        key: str
            The key associated with the item to be retrieved.

        Returns
        -------
        list
            The list of the items, split by batch and ordered.

        Notes
        -----
        An example of how the ordering is performed: Suppose the storage is initialized
        with `rollout=3` and `workers=2`. Then in each sampling iteration of the agent,
        a batch of 2 items (one for each worker) will be appended to storage for a total
        of 3 iterations. Thus, at the end of the iterations, the storage value for any given item
        will look similar to this::

            [[worker 1 sample 1, worker 2 sample 1],
             [worker 1 sample 2, worker 2 sample 2],
             [worker 1 sample 3, worker 2 sample 3]].

        The list returned by :meth:`~Storage.order` would then be::

            [worker 1 sample 1, worker 1 sample 2, worker 1 sample 3,
            worker 2 sample 1, worker 2 sample 2, worker 3 sample 3].

        The purpose of this ordering is to maintain chronologically consecutive blocks of rollout samples,
        which is useful for agents with recursive components.
        """
        if torch.is_tensor(self.storage[key][0]):
            ordered = torch.stack(self.storage[key][:self.rollout], -2)
            shape = ordered.shape
            ordered = ordered.view(*shape[:-3], self.rollout * self.workers, shape[-1])
            return ordered
        else:
            ordered = []
            for i in range(self.workers):
                ordered += [self.storage[key][j][i] for j in range(self.rollout)]
            return ordered


    def reset(self) -> None:
        """Empties the storage.

        """
        self.storage = {}


from .environment_wrapper import Task
from .conformer_env import ConformerEnv
import inspect

from conformer_rl.environments import environments

from gym.envs.registration import register

envs = [m[0] for m in inspect.getmembers(environments, inspect.isclass) if m[1].__module__ == 'conformer_rl.environments.environments']
for env in envs:
    register(
        id = env + '-v0',
        entry_point = 'conformer_rl.environments.environments:' + env
    )


import pathlib
from setuptools import setup
import setuptools

README = (pathlib.Path(__file__).parent / "README.md").read_text()

setuptools.setup(
    name="conformer-rl",
    version="0.1.1",
    description="Deep Reinforcement Library for Conformer Generation",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ZimmermanGroup/conformer-rl",
    author="Runxuan Jiang",
    author_email="runxuanj@umich.edu",
    license="MIT",
    package_dir={"": "src"},
    packages = setuptools.find_packages(where="src"),
    python_requires=">=3.7",
    install_requires=(
        'numpy',
        'tensorboard',
        'gym',
        'stable-baselines3>=1.0',
        'common_wrangler == 0.3.5',
        'ligninkmc',
        'stk',
        'py3Dmol',
        'seaborn',
        'jupyterlab'
    ),
    extras_require={
        "dev": ["sphinx", "sphinx_rtd_theme", "pytest", "coverage", "pytest-mock"]
    },
)


"""
RTGN_GAT
========
"""
from torch import nn
import torch
from torch_geometric.data import Data, Batch
import torch_geometric.nn as gnn

import numpy as np

from typing import List, Tuple, Dict

from conformer_rl.models.graph_components import GAT
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class RTGNGat(torch.nn.Module):
    """Actor-critic neural network using graph transformer network (GAT) [1]_
    for predicting discrete torsion angles.

    Works with molecules with any number of torsion angles, and batches containing graphs of different molecules.

    Parameters
    ----------
    action_dim : int
        The number of discrete action choices for each torsion angle.
    hidden_dim : int
        Dimension of the hidden layer.
    node_dim : int
        The dimension of each node embedding in the input graph.

    References
    ----------
    .. [1] `GAT paper <https://arxiv.org/abs/1911.06455>`_
    """
    def __init__(self, action_dim: int, hidden_dim: int, node_dim: int):
        super().__init__()
        self.action_dim = action_dim
        self.hidden_dim = hidden_dim

        self.actor = _RTGNGatActor(action_dim, hidden_dim, node_dim=node_dim)
        self.critic = _RTGNGatCritic(action_dim, hidden_dim, node_dim=node_dim)

    def forward(self, 
    obs: List[Tuple[Batch, List[List[int]]]],
    action: torch.Tensor=None) -> dict:
        """
        Parameters
        ----------
        obs : list of 2-tuples of Pytorch Geometric Batch objects and list of lists of int
            Each tuple is a single observation (the entire list is a batch). Each Pytorch Geometric Batch object corresponds to
            the Pytorch Geometric graph representing the molecule. The list of lists of integers
            is a list of all the torsions of the molecule, where each torsion is represented by a list of four integers, where the integers
            are the indices of the four atoms making up the torsion.
        action : batch of torch.Tensor, optional
            If specified, the log probabilities returned by the network will be the log probabilities for the specified
            actions instead of for the newly sampled actions.

        Returns
        -------
        prediction : dict
            * prediction['v']: The value estimation.
            * prediction['a']: The action sampled from the distribution predicted by the network.
            * prediction['entropy'] The entropy of the distribution.
            * prediction['log_pi_a'] The log probabilities of the actions from the distribution.
        """
        data_list = []
        nr_list = []
        for b, nr in obs:
            data_list += b.to_data_list()
            nr_list.append(torch.LongTensor(nr))

        data = Batch.from_data_list(data_list)
        data = data.to(device)
        N = data.num_graphs

        so_far = 0
        torsion_batch_idx = []
        torsion_list_sizes = []

        for i in range(N):
            nr_list[i] += so_far
            so_far += int((data.batch == i).sum())
            torsion_batch_idx.extend([i]*int(nr_list[i].shape[0]))
            torsion_list_sizes += [nr_list[i].shape[0]]

        nrs = torch.cat(nr_list).to(device)
        torsion_batch_idx = torch.LongTensor(torsion_batch_idx).to(device)
        obs = (data, nrs, torsion_batch_idx, torsion_list_sizes)

        logits = self.actor(obs)
        v = self.critic(obs)

        dist = torch.distributions.Categorical(logits=logits)
        if action == None:
            action = dist.sample()

        tls_max = np.array(torsion_list_sizes).max()
        log_prob = dist.log_prob(action[:,:tls_max])

        entropy = dist.entropy()

        prediction = {
            'a': action,
            'log_pi_a': log_prob,
            'ent': entropy,
            'v': v,
        }

        return prediction

class _RTGNGatCritic(torch.nn.Module):
    def __init__(self, action_dim, hidden_dim, node_dim):
        super().__init__()
        self.gat = GAT(hidden_dim=hidden_dim, node_dim=node_dim)
        self.set2set = gnn.Set2Set(hidden_dim, processing_steps=6)
        self.mlp = nn.Sequential(
            nn.Linear(2*hidden_dim, hidden_dim), 
            nn.ReLU(), 
            nn.Linear(hidden_dim, hidden_dim), 
            nn.ReLU(), 
            nn.Linear(hidden_dim, 1)
            )

    def forward(self, obs):
        data, nonring, nrbidx, torsion_list_sizes = obs
        N = data.num_graphs

        out = self.gat(data)
        pool = self.set2set(out, data.batch)
        v = self.mlp(pool)

        return v

class _RTGNGatActor(torch.nn.Module):
    def __init__(self, action_dim, hidden_dim, node_dim):
        super().__init__()
        self.gat = self.gat = GAT(hidden_dim=hidden_dim, node_dim=node_dim)
        self.set2set = gnn.Set2Set(hidden_dim, processing_steps=6)

        self.fc = nn.Linear(2*hidden_dim, hidden_dim)
        self.mlp = nn.Sequential(nn.Linear(5*hidden_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, action_dim))

        self.hidden_dim = hidden_dim

    def forward(self, obs):
        data, nonring, nrbidx, torsion_list_sizes = obs
        N = data.num_graphs

        out = self.gat(data)
        pool = self.set2set(out, data.batch)
        graph_embed = self.fc(pool)

        graph_embed = torch.index_select(
            graph_embed,
            dim=0,
            index=nrbidx
        )

        graph_embed = graph_embed.view(-1, self.hidden_dim)

        out = torch.index_select(
            out,
            dim=0,
            index=nonring.view(-1)
        )
        out = out.view(-1, self.hidden_dim * 4)
        out = torch.cat([graph_embed,out],1)   # shape = (num_torsions, 5*self.hidden_dim)
        out = self.mlp(out)

        logit = out.split(torsion_list_sizes)
        logit = torch.nn.utils.rnn.pad_sequence(logit).permute(1, 0, 2)

        return logit



"""
Observation_mixins
==================

Pre-built observation handlers.
"""
import numpy as np
import torch

from rdkit import Chem

from torch_geometric.data import Data, Batch
from torch_geometric.transforms import Distance, NormalizeScale, Center, NormalizeRotation
from conformer_rl.environments.environment_components import molecule_features

from typing import List, Tuple

class GraphObsMixin:
    """Represents molecule as a PyTorch Geometric graph where no information is included for each (node) and each edge
    is represented by a 6-tuple as generated by :func:`~conformer_rl.environments.environment_components.molecule_features.bond_type`.
    """
    def _obs(self) -> Tuple[Batch, List[List[int]]]:
        """
        returns
        -------
        Tuple[Batch, List[List[int]]
            The Batch object contains the Pytorch Geometric graph representing the molecule. The list of lists of integers
            is a list of all the torsions of the molecule, where each torsion is represented by a list of four integers, where the integers
            are the indices of the four atoms making up the torsion.
        """

        mol = Chem.rdmolops.RemoveHs(self.mol)
        conf = mol.GetConformer()
        atoms = mol.GetAtoms()
        bonds = mol.GetBonds()

        node_features = [[] for atom in atoms]
        edge_indices = molecule_features.get_bond_pairs(mol)
        edge_attributes = []
        for bond in bonds:
            edge_attributes += [molecule_features.bond_type(bond), molecule_features.bond_type(bond)]

        data = Data(
                    x=torch.tensor(node_features, dtype=torch.float),
                    edge_index=torch.tensor(edge_indices, dtype=torch.long),
                    edge_attr=torch.tensor(edge_attributes,dtype=torch.float),
                    pos=torch.Tensor(conf.GetPositions())
                )

        data = Center()(data)
        data = NormalizeRotation()(data)
        data.x = data.pos
        data = Batch.from_data_list([data])
        return data, self.nonring

class AtomTypeGraphObsMixin:
    """Represents molecule as a PyTorch Geometric graph where each node contains information about the atom's element.
    """
    def _obs(self) -> Tuple[Batch, List[List[int]]]:
        """
        returns
        -------
        Tuple[Batch, List[List[int]]
            The Batch object contains the Pytorch Geometric graph representing the molecule. The list of lists of integers
            is a list of all the torsions of the molecule, where each torsion is represented by a list of four integers, where the integers
            are the indices of the four atoms making up the torsion.
        """
        mol = Chem.rdmolops.RemoveHs(self.mol)
        conf = mol.GetConformer()
        atoms = mol.GetAtoms()
        bonds = mol.GetBonds()

        node_features = [molecule_features.atom_type_CO(atom) for atom in atoms]
        edge_indices = molecule_features.get_bond_pairs(mol)
        edge_attributes = []
        for bond in bonds:
            edge_attributes += [molecule_features.bond_type(bond), molecule_features.bond_type(bond)]

        data = Data(
                    x=torch.tensor(node_features, dtype=torch.float),
                    edge_index=torch.tensor(edge_indices, dtype=torch.long),
                    edge_attr=torch.tensor(edge_attributes,dtype=torch.float),
                    pos=torch.Tensor(conf.GetPositions())
                )

        data = Center()(data)
        data = NormalizeRotation()(data)
        data = Batch.from_data_list([data])
        return data, self.nonring


class AtomCoordsTypeGraphObsMixin:
    """Represents molecule as a PyTorch Geometric graph where each node contains information about the atom's element and three-dimensional coordinates.
    """
    def _obs(self) -> Tuple[Batch, List[List[int]]]:
        """
        returns
        -------
        Tuple[Batch, List[List[int]]
            The Batch object contains the Pytorch Geometric graph representing the molecule. The list of lists of integers
            is a list of all the torsions of the molecule, where each torsion is represented by a list of four integers, where the integers
            are the indices of the four atoms making up the torsion.
        """
        mol = Chem.rdmolops.RemoveHs(self.mol)
        conf = mol.GetConformer()
        atoms = mol.GetAtoms()
        bonds = mol.GetBonds()

        node_features = [molecule_features.atom_type_CO(atom) + molecule_features.atom_coords(atom, conf) for atom in atoms]
        edge_indices = molecule_features.get_bond_pairs(mol)
        edge_attributes = []
        for bond in bonds:
            edge_attributes += [molecule_features.bond_type(bond), molecule_features.bond_type(bond)]

        data = Data(
                    x=torch.tensor(node_features, dtype=torch.float),
                    edge_index=torch.tensor(edge_indices, dtype=torch.long),
                    edge_attr=torch.tensor(edge_attributes,dtype=torch.float),
                    pos=torch.Tensor(conf.GetPositions())
                )

        data = Center()(data)
        data = NormalizeRotation()(data)
        data.x[:,-3:] = data.pos
        data = Batch.from_data_list([data])
        return data, self.nonring


"""
Base_agent
==========
"""
import torch
import time
import logging
import numpy as np
from conformer_rl.utils import current_time, load_model, save_model, mkdir, to_np
from conformer_rl.logging import TrainLogger, EnvLogger
from conformer_rl.agents.storage import Storage
from conformer_rl.config import Config
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class BaseAgent:
    """ Base interface for building reinforcement learning agents.

    Parameters
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object for the agent. See notes for a list of config
        parameters used by specific pre-built agents.

    Attributes
    ----------
    config : :class:`~conformer_rl.config.agent_config.Config`
        Configuration object used for the agent.
    unique_tag : str
        Unique identifier string for the current training session. Used to identify
        this session in logging.
    eval_logger : :class:`~conformer_rl.logging.env_logger.EnvLogger`
        Used for logging environment data when evaluating agent.
    train_logger :  :class:`~conformer_rl.logging.train_logger.TrainLogger`
        Used for logging agent data while training.
    total_steps : int
        Total number of environment interactions/steps taken by the agent.
    storage : :class:`~conformer_rl.agents.storage.Storage`
        Used to save environment samples, analogous to a replay buffer.

    """
    def __init__(self, config: Config):
        self.config = config
        self.task = config.train_env # gym environment wrapper

        self.dir = config.data_dir
        self.unique_tag = f'{config.tag}_{current_time()}'

        self.eval_logger = EnvLogger(tag=self.unique_tag, dir=self.dir)
        self.train_logger = TrainLogger(tag=self.unique_tag, dir=self.dir, use_tensorboard=config.use_tensorboard, use_cache=False, use_print=False)
        self.total_steps = 0
        self.storage = Storage(config.rollout_length, config.num_workers)

    def run_steps(self) -> None:
        """ Trains the agent.

        Trains the agent until the maximum number of steps (specified by config) is reached.
        Also periodically saves neural network parameters and performs evaluations on the agent,
        if specified in the config.
        """
        config = self.config

        while self.total_steps < config.max_steps:
            if config.save_interval > 0 and self.total_steps % config.save_interval == 0:
                path = self.dir + '/' + 'models' + '/' + self.unique_tag
                mkdir(path)
                self.save(path + '/' +  str(self.total_steps) + '.model')

            if config.eval_interval > 0 and self.total_steps % config.eval_interval == 0:
                eval_start = time.time()
                self.evaluate()
                logging.debug(f'Eval at step {self.total_steps}, eval duration: {time.time() - eval_start} seconds')

            step_start = time.time()
            logging.debug(f'Starting agent step {self.total_steps}')
            self.step()
            logging.debug(f'agent step completed in {time.time() - step_start} seconds')

        self.task.close()

    def step(self) -> None:
        """Performs one iteration of acquiring samples on the environment
        and then trains on the acquired samples.
        """
        raise NotImplementedError

    def _eval_episode(self) -> dict:
        """Evalutes the agent on a single episode of the evaluation environment.

        Returns
        -------
        dict
            Information from the evaluation environment to be logged by the
            `eval_logger`.
        """
        env = self.config.eval_env
        state = env.reset()
        info = None
        done = False

        with torch.no_grad():
            while not done:
                prediction = self.network(state)
                action = prediction['a']
                state, reward, done, info = env.step(to_np(action))
                self.eval_logger.log_step(info[0]['step_info'])
            return info[0]['episode_info']

    def evaluate(self) -> None:
        """Evaluates the agent on the evaluation environment.

        Information dict returned by the environment's :meth:`conformer_rl.environments.conformer_env.ConformerEnv.step` method
        is logged by the `eval_logger` and saved.
        """
        returns = []
        for ep in range(self.config.eval_episodes):
            ep_info = self._eval_episode()
            returns.append(ep_info["total_rewards"])
            
            self.eval_logger.log_episode(ep_info)
            path = f'agent_step_{self.total_steps}' + '/' + f'ep_{ep}'
            self.eval_logger.save_episode(path, save_molecules=True)
            self.train_logger.add_scalar('episodic_return_eval', np.mean(returns), self.total_steps)

    def load(self, filename: str) -> None:
        """Loads the neural network with weights.

        Parameters
        ----------
        filename : str
            The path where the neural network weights are saved.
        """
        load_model(self.network, filename)

    def save(self, filename: str) -> None:
        """ Saves the neural network weights to a file.

        Parameters
        ----------
        filename : str
            The path where the neural network weights are to be saved.
        """
        save_model(self.network, filename)